import Foundation
import CryptoKit
import AppKit

typealias SourceFile = (data: Data, didChange: Bool)
typealias SourceTextFile = (content: String, didChange: Bool)

final class FileSystem {

    private static let tempFileName = "temp.bin"

    private static let hashesFileName = "hashes.json"

    private let input: URL

    private let output: URL

    private let source = "FileChangeMonitor"

    private var hashesFile: URL {
        input.appendingPathComponent(FileSystem.hashesFileName)
    }

    private var tempFile: URL {
        input.appendingPathComponent(FileSystem.tempFileName)
    }

    /**
     The hashes of all accessed files from the previous run

     The key is the relative path to the file from the source
     */
    private var previousFiles: [String : Data] = [:]

    /**
     The paths of all files which were accessed, with their new hashes

     This list is used to check if a file was modified, and to write all accessed files back to disk
     */
    private var accessedFiles: [String : Data] = [:]

    /**
     All files which should be copied to the output folder
     */
    private var requiredFiles: Set<String> = []

    /**
     The files marked as external in element metadata.

     Files included here are not generated, since they are assumed to be added separately.
     */
    private var externalFiles: Set<String> = []

    /**
     The files marked as expected, i.e. they exist after the generation is completed.

     The key of the dictionary is the file path, the value is the file providing the link
     */
    private var expectedFiles: [String : String] = [:]

    /**
     All pages without content which have been created
     */
    private var emptyPages: Set<String> = []

    /**
     All pages which have `status` set to ``PageState.draft``
     */
    private var draftPages: Set<String> = []

    /**
     All paths to page element folders, indexed by their unique id.

     This relation is used to generate relative links to pages using the ``Element.id`
     */
    private var pagePaths: [String: String] = [:]

    /**
     The image creation tasks.

     The key is the destination path.
     */
    private var imageTasks: [String : ImageOutput] = [:]

    /**
     The paths to all pages which were changed
     */
    private var generatedPages: Set<String> = []

    init(in input: URL, to output: URL) {
        self.input = input
        self.output = output

        guard exists(hashesFile) else {
            log.add(info: "No file hashes loaded, regarding all content as new", source: source)
            return
        }
        let data: Data
        do {
            data = try Data(contentsOf: hashesFile)
        } catch {
            log.add(
                warning: "File hashes could not be read, regarding all content as new",
                source: source,
                error: error)
            return
        }
        do {
            self.previousFiles = try JSONDecoder().decode(from: data)
        } catch {
            log.add(
                warning: "File hashes could not be decoded, regarding all content as new",
                source: source,
                error: error)
            return
        }
    }

    func urlInOutputFolder(_ path: String) -> URL {
        output.appendingPathComponent(path)
    }

    func urlInContentFolder(_ path: String) -> URL {
        input.appendingPathComponent(path)
    }

    /**
     Get the current hash of file data at a path.

     If the hash has been computed previously during the current run, then this function directly returns it.
     */
    private func hash(_ data: Data, at path: String) -> Data {
        accessedFiles[path] ?? SHA256.hash(data: data).data
    }

    private func exists(_ url: URL) -> Bool {
        FileManager.default.fileExists(atPath: url.path)
    }

    func dataOfRequiredFile(atPath path: String, source: String) -> Data? {
        let url = input.appendingPathComponent(path)
        guard exists(url) else {
            log.failedToOpen(path, requiredBy: source, error: nil)
            return nil
        }

        do {
            return try Data(contentsOf: url)
        } catch {
            log.failedToOpen(path, requiredBy: source, error: error)
            return nil
        }
    }

    func dataOfOptionalFile(atPath path: String, source: String) -> Data? {
        let url = input.appendingPathComponent(path)
        guard exists(url) else {
            return nil
        }

        do {
            return try Data(contentsOf: url)
        } catch {
            log.failedToOpen(path, requiredBy: source, error: error)
            return nil
        }
    }

    func contentOfOptionalFile(atPath path: String, source: String, createEmptyFileIfMissing: Bool = false) -> String? {
        let url = input.appendingPathComponent(path)
        guard exists(url) else {
            if createEmptyFileIfMissing {
                try? Data().write(to: url)
            }
            return nil
        }

        do {
            return try String(contentsOf: url)
        } catch {
            log.failedToOpen(path, requiredBy: source, error: error)
            return nil
        }
    }

    private func getData(atPath path: String) -> SourceFile? {
        let url = input.appendingPathComponent(path)
        guard exists(url) else {
            return nil
        }
        let data: Data
        do {
            data = try Data(contentsOf: url)
        } catch {
            log.add(error: "Failed to read data at \(path)", source: source, error: error)
            return nil
        }
        let newHash = hash(data, at: path)
        defer {
            accessedFiles[path] = newHash
        }
        guard let oldHash = previousFiles[path] else {
            return (data: data, didChange: true)
        }
        return (data: data, didChange: oldHash != newHash)
    }

    func writeHashes() {
        do {
            let encoder = JSONEncoder()
            encoder.outputFormatting = .prettyPrinted
            let data = try encoder.encode(accessedFiles)
            try data.write(to: hashesFile)
        } catch {
            log.add(warning: "Failed to save file hashes", source: source, error: error)
        }
    }

    // MARK: Images

    private func loadImage(atPath path: String) -> (image: NSImage, changed: Bool)? {
        guard let (data, changed) = getData(atPath: path) else {
            log.add(error: "Failed to load file", source: path)
            return nil
        }
        guard let image = NSImage(data: data) else {
            log.add(error: "Failed to read image", source: path)
            return nil
        }
        return (image, changed)
    }

    @discardableResult
    func requireImage(source: String, destination: String, width: Int, desiredHeight: Int? = nil)  -> NSSize {
        let height = desiredHeight.unwrapped(CGFloat.init)
        let sourceUrl = input.appendingPathComponent(source)
        let image = ImageOutput(source: source, width: width, desiredHeight: desiredHeight)

        let standardSize = NSSize(width: CGFloat(width), height: height ?? CGFloat(width) / 16 * 9)
        guard sourceUrl.exists else {
            log.add(error: "Missing file with size (\(width),\(desiredHeight ?? -1))",
                    source: source)
            return standardSize
        }
        guard let imageSize = loadImage(atPath: image.source)?.image.size else {
            log.add(error: "Unreadable image with size (\(width),\(desiredHeight ?? -1))",
                    source: source)
            return standardSize
        }
        let scaledSize = imageSize.scaledDown(to: CGFloat(width))

        guard let existing = imageTasks[destination] else {
            imageTasks[destination] = image
            return scaledSize
        }
        guard existing.source == source else {
            log.add(error: "Multiple sources (\(existing.source),\(source))",
                    source: destination)
            return scaledSize
        }
        guard existing.hasSimilarRatio(as: image) else {
            log.add(error: "Multiple ratios (\(existing.ratio!),\(image.ratio!))",
                    source: destination)
            return scaledSize
        }
        if image.width > existing.width {
            log.add(info: "Increasing size from \(existing.width) to \(width)",
                    source: destination)
            imageTasks[destination] = image
        }
        return scaledSize
    }

    func createImages() {
        for (destination, image) in imageTasks.sorted(by: { $0.key < $1.key }) {
            createImageIfNeeded(image, for: destination)
        }
    }

    private func createImageIfNeeded(_ image: ImageOutput, for destination: String) {
        guard let (sourceImageData, sourceImageChanged) = getData(atPath: image.source) else {
            log.add(error: "Failed to open file", source: image.source)
            return
        }

        let destinationUrl = output.appendingPathComponent(destination)

        // Check if image needs to be updated
        guard !destinationUrl.exists || sourceImageChanged else {
            return
        }

        // Ensure that image file is supported
        let ext = destinationUrl.pathExtension.lowercased()
        guard ImageType(fileExtension: ext) != nil else {
            // TODO: This should never be reached, since extensions are checked before
            log.add(info: "Copying image", source: image.source)
            do {
                let sourceUrl = input.appendingPathComponent(image.source)
                try destinationUrl.ensureParentFolderExistence()
                try sourceUrl.copy(to: destinationUrl)
            } catch {
                log.add(error: "Failed to copy image", source: destination)
            }
            return
        }
        guard let sourceImage = NSImage(data: sourceImageData) else {
            log.add(error: "Failed to read file", source: image.source)
            return
        }

        let desiredWidth = CGFloat(image.width)
        let desiredHeight = image.desiredHeight.unwrapped(CGFloat.init)

        let destinationSize = sourceImage.size.scaledDown(to: desiredWidth)
        let scaledImage = sourceImage.scaledDown(to: destinationSize)
        let scaledSize = scaledImage.size

        if abs(scaledImage.size.width - desiredWidth) > 2 {
            log.add(warning: "Desired width \(desiredWidth), got \(scaledSize.width)", source: destination)
        }
        if abs(destinationSize.height - scaledImage.size.height) > 2 {
            log.add(warning: "Desired height \(destinationSize.height), got \(scaledSize.height)", source: destination)
        }
        if let desiredHeight = desiredHeight {
            let desiredRatio = desiredHeight / desiredWidth
            let adjustedDesiredHeight = scaledSize.width * desiredRatio
            if abs(adjustedDesiredHeight - scaledSize.height) > 5 {
                log.add(warning: "Desired height \(desiredHeight), got \(scaledSize.height)", source: destination)
                return
            }
        }
        if scaledSize.width > desiredWidth {
            log.add(warning:" Desired width \(desiredWidth), got \(scaledSize.width)", source: destination)
        }

        let destinationExtension = destinationUrl.pathExtension.lowercased()
        guard let type = ImageType(fileExtension: destinationExtension)?.fileType else {
            log.add(error: "No image type for extension \(destinationExtension)",
                    source: destination)
            return
        }
        guard let tiff = scaledImage.tiffRepresentation, let tiffData = NSBitmapImageRep(data: tiff) else {
            log.add(error: "Failed to get data", source: image.source)
            return
        }

        guard let data = tiffData.representation(using: type, properties: [.compressionFactor: NSNumber(0.7)]) else {
            log.add(error: "Failed to get data", source: image.source)
            return
        }
        do {
            try data.createFolderAndWrite(to: destinationUrl)
        } catch {
            log.add(error: "Failed to write image \(destination)", source: "Image Processor", error: error)
            return
        }
    }

    // MARK: File copying

    /**
     Add a file as required, so that it will be copied to the output directory.
     */
    func require(file: String) {
        let url = input.appendingPathComponent(file)
        guard url.exists, url.isDirectory else {
            requiredFiles.insert(file)
            return
        }
        do {
            try FileManager.default
                .contentsOfDirectory(atPath: url.path)
                .forEach {
                    // Recurse into subfolders
                    require(file: file + "/" + $0)
                }
        } catch {
            log.add(error: "Failed to read folder \(file): \(error)", source: source)
        }
    }

    /**
     Mark a file as explicitly missing.

     This is done for the `externalFiles` entries in metadata,
     to indicate that these files will be copied to the output folder manually.
     */
    func exclude(file: String) {
        externalFiles.insert(file)
    }

    /**
     Mark a file as expected to be present in the output folder after generation.

     This is done for all links between pages, which only exist after the pages have been generated.
     */
    func expect(file: String, source: String) {
        expectedFiles[file] = source
    }

    func copyRequiredFiles() {
        var copiedFiles = Set<String>()
        for file in requiredFiles {
            let cleanPath = cleanRelativeURL(file)
            let sourceUrl = input.appendingPathComponent(cleanPath)
            let destinationUrl = output.appendingPathComponent(cleanPath)
            guard sourceUrl.exists else {
                if !isExternal(file: file) {
                    log.add(error: "Missing required file", source: cleanPath)
                }
                continue
            }
            if copyFileIfChanged(from: sourceUrl, to: destinationUrl) {
                copiedFiles.insert(file)
            }
        }
        try? tempFile.delete()
        for (file, source) in expectedFiles {
            guard !isExternal(file: file) else {
                continue
            }
            let cleanPath = cleanRelativeURL(file)
            let destinationUrl = output.appendingPathComponent(cleanPath)
            if !destinationUrl.exists {
                log.add(error: "Missing \(cleanPath)", source: source)
            }
        }
        guard !copiedFiles.isEmpty else {
            print("No required files copied")
            return
        }
        print("\(copiedFiles.count) required files copied:")
        for file in copiedFiles.sorted() {
            print("  " + file)
        }
    }

    private func copyFileIfChanged(from sourceUrl: URL, to destinationUrl: URL) -> Bool {
        guard configuration.minifyCSSandJS else {
            return copyBinaryFileIfChanged(from: sourceUrl, to: destinationUrl)
        }
        switch sourceUrl.pathExtension.lowercased() {
        case "js":
            return minifyJS(at: sourceUrl, andWriteTo: destinationUrl)
        case "css":
            return minifyCSS(at: sourceUrl, andWriteTo: destinationUrl)
        default:
            return copyBinaryFileIfChanged(from: sourceUrl, to: destinationUrl)
        }
    }

    private func copyBinaryFileIfChanged(from sourceUrl: URL, to destinationUrl: URL) -> Bool {
        do {
            let data = try Data(contentsOf: sourceUrl)
            return writeIfChanged(data, to: destinationUrl)
        } catch {
            log.add(error: "Failed to read data at \(sourceUrl.path)", source: source, error: error)
            return false
        }
    }

    private func minifyJS(at sourceUrl: URL, andWriteTo destinationUrl: URL) -> Bool {
        let command = "uglifyjs \(sourceUrl.path) > \(tempFile.path)"
        do {
            _ = try safeShell(command)
            return copyBinaryFileIfChanged(from: tempFile, to: destinationUrl)
        } catch {
            log.add(error: "Failed to minify \(sourceUrl.path): \(error)", source: source)
            return false
        }
    }

    private func minifyCSS(at sourceUrl: URL, andWriteTo destinationUrl: URL) -> Bool {
        let command = "cleancss \(sourceUrl.path) -o \(tempFile.path)"
        do {
            _ = try safeShell(command)
            return copyBinaryFileIfChanged(from: tempFile, to: destinationUrl)
        } catch {
            log.add(error: "Failed to minify \(sourceUrl.path): \(error)", source: source)
            return false
        }
    }

    private func cleanRelativeURL(_ raw: String) -> String {
        let raw = raw.dropAfterLast("#") // Clean links to page content
        guard raw.contains("..") else {
            return raw
        }
        var result: [String] = []
        for component in raw.components(separatedBy: "/") {
            if component == ".." {
                _ = result.popLast()
            } else {
                result.append(component)
            }
        }
        return result.joined(separator: "/")
    }

    /**
     Check if a file is marked as external.

     Also checks for sub-paths of the file, e.g if the folder `docs` is marked as external,
     then files like `docs/index.html` are also found to be external.
     - Note: All paths are either relative to root (no leading slash) or absolute paths of the domain (leading slash)
     */
    func isExternal(file: String) -> Bool {
        // Deconstruct file path
        var path = ""
        for part in file.components(separatedBy: "/") {
            guard part != "" else {
                continue
            }
            if path == "" {
                path = part
            } else {
                path += "/" + part
            }
            if externalFiles.contains(path) {
                return true
            }
        }
        return false
    }

    func printExternalFiles() {
        guard !externalFiles.isEmpty else {
            return
        }
        print("\(externalFiles.count) external resources needed:")
        for file in externalFiles.sorted() {
            print("  " + file)
        }
    }

    // MARK: Pages

    func isEmpty(page: String) {
        emptyPages.insert(page)
    }

    func printEmptyPages() {
        guard !emptyPages.isEmpty else {
            return
        }
        print("\(emptyPages.count) empty pages:")
        for page in emptyPages.sorted() {
            print("  " + page)
        }
    }

    func isDraft(path: String) {
        draftPages.insert(path)
    }

    func printDraftPages() {
        guard !draftPages.isEmpty else {
            return
        }
        print("\(draftPages.count) drafts:")
        for page in draftPages.sorted() {
            print("  " + page)
        }
    }

    func add(page: String, id: String) {
        if let existing = pagePaths[id] {
            log.add(error: "Conflicting id with \(existing)", source: page)
        }
        pagePaths[id] = page
    }

    func getPage(for id: String) -> String? {
        pagePaths[id]
    }

    func generated(page: String) {
        generatedPages.insert(page)
    }

    func printGeneratedPages() {
        guard !generatedPages.isEmpty else {
            print("No pages modified")
            return
        }
        print("\(generatedPages.count) pages modified")
        for page in generatedPages.sorted() {
            print("  " + page)
        }
    }


    // MARK: Writing files

    @discardableResult
    func writeIfChanged(_ data: Data, to url: URL) -> Bool {
        // Only write changed files
        if url.exists, let oldContent = try? Data(contentsOf: url), data == oldContent {
            return false
        }
        do {
            try data.createFolderAndWrite(to: url)
            return true
        } catch {
            log.add(error: "Failed to write file", source: url.path, error: error)
            return false
        }
    }

    @discardableResult
    func write(_ string: String, to url: URL) -> Bool {
        let data = string.data(using: .utf8)!
        return writeIfChanged(data, to: url)
    }

    // MARK: Running other tasks

    @discardableResult
    func safeShell(_ command: String) throws -> String {
        let task = Process()
        let pipe = Pipe()

        task.standardOutput = pipe
        task.standardError = pipe
        task.arguments = ["-cl", command]
        task.executableURL = URL(fileURLWithPath: "/bin/zsh")
        task.standardInput = nil

        try task.run()

        let data = pipe.fileHandleForReading.readDataToEndOfFile()
        let output = String(data: data, encoding: .utf8)!

        return output
    }
}

private extension Digest {

    var bytes: [UInt8] { Array(makeIterator()) }

    var data: Data { Data(bytes) }

    var hexStr: String {
        bytes.map { String(format: "%02X", $0) }.joined()
    }
}
